//
// Authors:
//   Patrick Cernko <errror@ps.uni-sb.de>
//
// Copyright:
//   Patrick Cernko, 2004
//
// Last Change:
//   $Date: 2004/01/21 17:51:22 $ by $Author: errror $
//   $Revision: 1.2 $
//

#if defined(INTERFACE)
#pragma implementation "AEEvalInterpreter.hh"
#endif

#include "AEEvalInterpreter.hh"

#include "AELanguageLayer.hh"
#include "AEEvalFrame.hh"
#include "AEData.hh"

#include <cstdio>   // fprintf()


////////////////////////////////////////////////////////////////////////////////
// AEEvalInterpreter
////////////////////////////////////////////////////////////////////////////////
// initializing of the static members
AEEvalInterpreter* AEEvalInterpreter::self;

void AEEvalInterpreter::Init() {
  // create a Interpreter and store it in the "singleton" variable
  self = new AEEvalInterpreter();
  // XXX: what if self already contains a interpreter, who makes the gc?
}

void AEEvalInterpreter::PushCall(Closure*) {
  // do I need this?
//   AEClosure* closure = STATIC_CAST(AEClosure *, closure0);
//   AEEvalFrame::New(self, closure->GetRecord(), closure->GetLabel());
}

u_int AEEvalInterpreter::GetFrameSize(StackFrame* sFrame) {
  // make the StackFrame a AEEvalFrame
  AEEvalFrame* frame = STATIC_CAST(AEEvalFrame *, sFrame);
  // assertion: this frame must assoziated with US!
  Assert(sFrame->GetWorker() == this);
  // return it's size
  return frame->GetSize();
}

Worker::Result AEEvalInterpreter::Run(StackFrame* sFrame) {
  // make the StackFrame a AEEvalFrame
  AEEvalFrame* frame = STATIC_CAST(AEEvalFrame *, sFrame);
  // assertion: this frame must assoziated with US!
  Assert(sFrame->GetWorker() == this);

  if (AELanguageLayer::verbosity > 1) { fprintf(stderr, "Have working tree? ", frame); }
  AENode* current_tree = frame->GetNode();
  if (current_tree == INVALID_POINTER) {
    if (AELanguageLayer::verbosity > 1) { fprintf(stderr, "No, getting current tree from Scheduler.\n"); }
    current_tree = AENode::FromWord(Scheduler::currentArgs[0]);
    if (current_tree == INVALID_POINTER) {
      // fine, we do not have a current_tree and the Scheduler also
      // does not have a tree, but a number:
      Scheduler::PopFrame(frame->GetSize());
      if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Current Result: %d\n", Store::WordToInt(Scheduler::currentArgs[0])); }
      // leave current result on Scheduler
      return Worker::CONTINUE;
    } else {
      if (AELanguageLayer::verbosity > 2) { current_tree->Dump(0); }
      frame->SetNode(current_tree);
    }
    Scheduler::nArgs = 0;
  } else {
    if (AELanguageLayer::verbosity > 1) { fprintf(stderr, "Yes!\n"); }
    if (AELanguageLayer::verbosity > 2) { current_tree->Dump(0); }
  }


  // we have a current_tree
  // check if we also already have a left result
  if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Left side reduced? "); }
  s_int left_result = current_tree->GetLeftNumber();
  if (left_result == INVALID_INT) {
    if (Scheduler::nArgs > 0) {
      // the intermediate result must be on the Scheduler's args
      if (AELanguageLayer::verbosity > 0) {  fprintf(stderr, "No, but Scheduler has: "); }
      left_result = Store::WordToInt(Scheduler::currentArgs[0]);
      Scheduler::nArgs = 0;
      if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "%d\n", left_result); }
      // store it as new left subtree
      current_tree = AENode::New(current_tree->GetArithmeticOperation(),
                                 Scheduler::currentArgs[0], current_tree->GetRight());
      frame->SetNode(current_tree);
    }

    else {
      // the scheduler just gave us a tree to evaluate and the left
      // tree is not yet done, let's do it!
      if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "No, evaluating left subtree.\n"); }
      if (AELanguageLayer::verbosity > 2) { current_tree->Dump(0); }
      // we have a real tree on the Scheduler's args but no current
      // one, so we have to compute the one on the Scheduler's args
      // first we save the tree in our frame
      frame->SetNode(current_tree);
      // now we can compute the left subtree
      Scheduler::nArgs = 1;
      Scheduler::currentArgs[0] = current_tree->GetLeft();
      // by creating a StackFrame to compute it
      AEEvalFrame::New(self);
      return Worker::CONTINUE;
    }
  } else {
    if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Yes: %d\n", left_result); }
  }


  // check if we also already have a right result
  if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Right side reduced? "); }
  s_int right_result = current_tree->GetRightNumber();
  if (right_result == INVALID_INT) {
    if (Scheduler::nArgs > 0) {
      // the intermediate result must be on the Scheduler's args
      if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "No, but Scheduler has: "); }
      right_result = Store::WordToInt(Scheduler::currentArgs[0]);
      Scheduler::nArgs = 0;
      if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "%d\n", right_result); }
    }

    else {
      if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "No, evaluating right subtree.\n"); }
      if (AELanguageLayer::verbosity > 2) { current_tree->Dump(0); }
      // and compute the right tree now
      Scheduler::nArgs = 1;
      Scheduler::currentArgs[0] = current_tree->GetRight();
      // by creating a StackFrame to compute it
      AEEvalFrame::New(self);
      return Worker::CONTINUE;
    }
  } else {
    if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Yes: %d\n", right_result); }
  }


  // yes, we have both results
  // so we can compute the result for our tree:
  s_int result(0);
  switch (current_tree->GetArithmeticOperation()) {
  case AENode::PLUS:
    if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Reducing: %d + %d = ", left_result, right_result); }
    result = left_result + right_result;
    break;
  case AENode::MINUS:
    if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Reducing: %d - %d = ", left_result, right_result); }
    result = left_result - right_result;
    break;
  case AENode::MUL:
    if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Reducing: %d * %d = ", left_result, right_result); }
    result = left_result * right_result;
    break;
  case AENode::DIV:
    if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Reducing: %d / %d = ", left_result, right_result); }
    result = left_result / right_result;
    break;
  case AENode::INVALID_OPERATION:
    fprintf(stderr, "current AENode has INVALID_OPERATION set!\n");
    return Worker::EXIT;
    break;
  }

  if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "%d\n", result); }
  Scheduler::nArgs = 1;
  Scheduler::currentArgs[0] = Store::IntToWord(result);
  Scheduler::PopFrame(frame->GetSize());

  return Worker::CONTINUE;
}

u_int AEEvalInterpreter::GetInArity(ConcreteCode*) {
  // only dummy
  return 0;
}

u_int AEEvalInterpreter::GetOutArity(ConcreteCode*) {
  // only dummy
  return 1;
}

const char* AEEvalInterpreter::Identify() {
  // yes, we are a ...
  return "AEEvalInterpreter";
}

void AEEvalInterpreter::DumpFrame(StackFrame* // sFrame
                                  ) {
//   // make the StackFrame a AEEvalFrame
//   AEEvalFrame* frame = STATIC_CAST(AEEvalFrame *, sFrame);
//   // assertion: this frame must assoziated with US!
//   Assert(sFrame->GetWorker() == this);

//   // ok, what can we say about the frame? Hmm, only its ae, I think!
//   std::fprintf(stderr, "Select %s\n", frame->GetAE()->ExportC());
}
