//
// Authors:
//   Patrick Cernko <errror@ps.uni-sb.de>
//
// Copyright:
//   Patrick Cernko, 2004
//
// Last Change:
//   $Date: 2004/01/21 17:51:22 $ by $Author: errror $
//   $Revision: 1.12 $
//
// Based on LazySelInterpreter.cc from alice-on-seam
// Authors:
//   Thorsten Brunklaus <brunklaus@ps.uni-sb.de>
//   Leif Kornstaedt <kornstae@ps.uni-sb.de>
//
// Copyright:
//   Thorsten Brunklaus, 2002
//   Leif Kornstaedt, 2002
//

#if defined(INTERFACE)
#pragma implementation "AEInterpreter.hh"
#endif

#include "AEInterpreter.hh"

#include "AELanguageLayer.hh"
#include "AEFrame.hh"
#include "AEData.hh"

#include <cstdio>   // fprintf()
#include <ctype.h>  // isdigit()
#include <errno.h>  // errno, ERANGE
#include <limits.h> // ULONG_MAX


////////////////////////////////////////////////////////////////////////////////
// AEInterpreter
////////////////////////////////////////////////////////////////////////////////
// initializing of the static members
AEInterpreter* AEInterpreter::self;

void AEInterpreter::Init() {
  // create a Interpreter and store it in the "singleton" variable
  self = new AEInterpreter();
  // XXX: what if self already contains a interpreter, who makes the gc?
}

void AEInterpreter::PushCall(Closure*) {
  // do I need this?
//   AEClosure* closure = STATIC_CAST(AEClosure *, closure0);
//   AEFrame::New(self, closure->GetRecord(), closure->GetLabel());
}

u_int AEInterpreter::GetFrameSize(StackFrame* sFrame) {
  // make the StackFrame a AEFrame
  AEFrame* frame = STATIC_CAST(AEFrame *, sFrame);
  // assertion: this frame must assoziated with US!
  Assert(sFrame->GetWorker() == this);
  // return it's size
  return frame->GetSize();
}

String* EatUpWhiteSpace(String* src) {
  u_char* cstring(src->GetValue());
  u_int pos(0);
  u_int size(src->GetSize());

  // eat up whitespace
  if (!isspace(cstring[pos])) {
    return src;
  } else {
    while (isspace(cstring[pos])) {
      pos++;
      if (pos >= size) {
        break;
      }
    }
    return String::New((const char*) cstring+pos, size-pos);
  }
}

Worker::Result AEInterpreter::Run(StackFrame* sFrame) {
  // make the StackFrame a AEFrame
  AEFrame* frame = STATIC_CAST(AEFrame *, sFrame);
  // assertion: this frame must assoziated with US!
  Assert(sFrame->GetWorker() == this);
  // get the arithmetic expression out of the frame
  String* ae = frame->GetAE();
  // ok, what's now? Interpret the String?
  if (ae->GetSize() == 0) {
    // treat empty expressions as illegal
    fprintf(stderr, "Illegal empty Expression!\n");
    return Worker::EXIT;
  }
  else {

    switch (frame->GetParsingAction()) {
    case AEFrame::PARSE_EXPRESSION:
      if (AELanguageLayer::verbosity > 1) { fprintf(stderr, "Interpreting Expression\n"); }

      if (Scheduler::nArgs == 0) {
        // if we do not have parsed the first term
        // emit frame for parsing term
        AEFrame::New(self, ae, AEFrame::PARSE_TERM);
      }

      else { // we have arguments on the Schedulers current ones

        // get the substring from the current Scheduler args
        word subtree = Scheduler::currentArgs[0];
        s_int last_op = frame->GetOperation();
        if (last_op != AENode::INVALID_OPERATION) {
          subtree = AENode::New((AENode::ArithmeticOperation) last_op,
                                frame->GetNode(), subtree)->ToWord();
        }
        frame->SetNode(subtree);
        // get the substring from the current Scheduler args
        ae = String::FromWordDirect(Scheduler::currentArgs[1]);
        Scheduler::nArgs = 0;

        ae = EatUpWhiteSpace(ae);
        u_char* ae_string(ae->GetValue());
        u_int ae_size(ae->GetSize());

        if (ae_size > 0) {
          // check wether it is a expression seperator sign
          if (ae_string[0] == '+') {
            if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "TOKEN: PLUS\n"); }
            frame->SetOperation(AENode::PLUS);
          } else if (ae_string[0] == '-') {
            if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "TOKEN: MINUS\n"); }
            frame->SetOperation(AENode::MINUS);
          } else if (ae_string[0] == ')') {
            if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "TOKEN: CLOSE_PAR\n"); }
            // build up new substring 
            String* subae = String::New((const char*) ae_string+1, ae_size-1);

            // put the results on the Scheduler
            Scheduler::nArgs = 2;
            Scheduler::currentArgs[0] = subtree;
            Scheduler::currentArgs[1] = subae->ToWord();
            
            // pop this frame
            Scheduler::PopFrame(frame->GetSize());

            return Worker::CONTINUE;
          } else {
            fprintf(stderr, "Illegal Expression seperator (+|-): \"%c\"\n", ae_string[0]);
            return Worker::EXIT;
          }
          // build up new substring for term parsing
          String* subae = String::New((const char*) ae_string+1, ae_size-1);
          // emit a new frame from parsing the term
          AEFrame::New(self, subae, AEFrame::PARSE_TERM);
        }
        else { // substring from parse_term is empty!
          // we are finished!
          if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "Parsed output!\n"); }
          if (AELanguageLayer::verbosity > 1) { AENode::Dump(subtree, 0); }

          Scheduler::nArgs = 2;
          Scheduler::currentArgs[0] = subtree;
          Scheduler::currentArgs[1] = ae->ToWord();

          Scheduler::PopFrame(frame->GetSize());
        }
      }
      break;

    case AEFrame::PARSE_TERM:
      if (AELanguageLayer::verbosity > 1) { fprintf(stderr, "Interpreting Term\n"); }

      if (Scheduler::nArgs == 0) {
        // if we do not have parsed the first number
        // emit frame for parsing number
        AEFrame::New(self, ae, AEFrame::PARSE_NUMBER);
      }

      else {

        // get the substring from the current Scheduler args
        word subtree = Scheduler::currentArgs[0];
        s_int last_op = frame->GetOperation();
        if (last_op != AENode::INVALID_OPERATION) {
          subtree = AENode::New((AENode::ArithmeticOperation) last_op,
                                frame->GetNode(), subtree)->ToWord();
        }
        frame->SetNode(subtree);
        // get the substring from the current Scheduler args
        ae = String::FromWordDirect(Scheduler::currentArgs[1]);
        Scheduler::nArgs = 0;

        ae = EatUpWhiteSpace(ae);
        u_char* ae_string(ae->GetValue());
        u_int ae_size(ae->GetSize());

        if (ae->GetSize() > 0) {
          // check wether it is a expression seperator sign
          if (ae_string[0] == '*') {
            if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "TOKEN: MUL\n"); }
            frame->SetOperation(AENode::MUL);
          } else if (ae_string[0] == '/') {
            if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "TOKEN: DIV\n"); }
            frame->SetOperation(AENode::DIV);
          } else {
            // this must be a expression seperator (+|-), so
            // pop this frame
            Scheduler::PopFrame(frame->GetSize());
            // and put the substring we got from parse_number
            // on the Scheduler's current args as is
            Scheduler::nArgs = 2;
            Scheduler::currentArgs[0] = subtree;
            Scheduler::currentArgs[1] = ae->ToWord();
            // now the "PARSE_EXPRESSION" frame should be on top!
            return Worker::CONTINUE;
          }
          // build up new substring for number parsing
          String* subae = String::New((const char*) ae_string+1, ae_size-1);
          // emit a new frame from parsing the number
          AEFrame::New(self, subae, AEFrame::PARSE_NUMBER);
        }

        else { // substring from parse_number is empty
          // pop this frame
          Scheduler::PopFrame(frame->GetSize());
          // and place the substring on the Scheduler's args again
          Scheduler::nArgs = 2;
          Scheduler::currentArgs[0] = (word) subtree;
          Scheduler::currentArgs[1] = ae->ToWord();
          // now the "PARSE_EXPRESSION" frame should be on top!
        }
      }
      break;

    case AEFrame::PARSE_NUMBER:
      if (AELanguageLayer::verbosity > 1) { fprintf(stderr, "Interpreting Number\n"); }

      if (Scheduler::nArgs == 0) {

        ae = EatUpWhiteSpace(ae);
        u_char* ae_string(ae->GetValue());
        u_int ae_pos(0);
        u_int ae_size(ae->GetSize());

        if (((ae_string[ae_pos] == '-') && isdigit(ae_string[ae_pos+1])) ||
            isdigit(ae_string[ae_pos])) {

          // build up digit-cstring
          char numcstr[ae_size-ae_pos+1];
          u_int pos(0);
          while (isdigit(ae_string[ae_pos])) {
            numcstr[pos] = ae_string[ae_pos];
            pos++;
            ae_pos++;
            if (ae_pos >= ae_size) {
              // end of ae-substring reached, no more digits
              break;
            }
          }

          // maybe assert pos < numcstr.length
          numcstr[pos] = '\0';
          s_int number = strtol(numcstr, NULL, 10);
          // make sure, number is small enough to be stored in a word
          if (((errno == ERANGE) &&
               ((number == LONG_MAX) || 
                (number == LONG_MIN))) ||
              ((number > MAX_VALID_INT) ||
               (number < MIN_VALID_INT))) {
            fprintf(stderr, "Integer out of range: %d\n", number);
            return Worker::EXIT;
          }
          if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "TOKEN: [%d]\n", number); }

        
          // build up expression subtree
          word subtree = Store::IntToWord(number);
          // build up substring from after the number
          word subae = String::New((const char*) ae_string+ae_pos,
                                   ae_size-ae_pos)->ToWord();
          // pop this frame
          Scheduler::PopFrame(frame->GetSize());
          // and put the new substring in the Scheduler's current args
          Scheduler::nArgs = 2;
          Scheduler::currentArgs[0] = subtree;
          Scheduler::currentArgs[1] = subae;
          // now the "PARSE_TERM" frame should be on top!
        }

        else if (ae_string[ae_pos] == '(') {
          if (AELanguageLayer::verbosity > 0) { fprintf(stderr, "TOKEN: OPEN_PAR\n"); }
          // build up new substring for expression parsing
          String* subae = String::New((const char*) ae_string+ae_pos+1, ae->GetSize()-ae_pos-1);
          // emit a new frame from parsing the number
          AEFrame::New(self, subae, AEFrame::PARSE_EXPRESSION);
        }

        else {
          // no number
          fprintf(stderr, "Parse error, not a number: \"%c\"\n", ae_string[ae_pos]);
          return Worker::EXIT;
        }
      }
      else {
        Scheduler::PopFrame(frame->GetSize());
      }
      // now the "PARSE_TERM" frame should be on top!
      break;

    } // end switch

  } // end else: ae.size != 0

  return Worker::CONTINUE;
}

u_int AEInterpreter::GetInArity(ConcreteCode*) {
  // only dummy
  return 0;
}

u_int AEInterpreter::GetOutArity(ConcreteCode*) {
  // only dummy
  return 1;
}

const char* AEInterpreter::Identify() {
  // yes, we are a ...
  return "AEInterpreter";
}

void AEInterpreter::DumpFrame(StackFrame* sFrame) {
  // make the StackFrame a AEFrame
  AEFrame* frame = STATIC_CAST(AEFrame *, sFrame);
  // assertion: this frame must assoziated with US!
  Assert(sFrame->GetWorker() == this);

  // ok, what can we say about the frame? Hmm, only its ae, I think!
  std::fprintf(stderr, "Select %s\n", frame->GetAE()->ExportC());
}
